@@ -121,6 +121,18 @@ class EnginePrint:
         del messages[:-1]
         return messages[-1]['content']
 
+class EnginePython:
+    def __init__(self, api_key = None):
+        pass
+
+
+    def prompt(self, messages, actors):
+        # remove all elements from the list except the last one
+        messages[-1]['content'] = messages[-2]['content']
+        del messages[:-1]
+        return messages[-1]['content']
+
+
 
 class NoMagic:
     """" This implementation uses exact string matching"""
@@ -137,7 +149,7 @@ class TrueMagic:
 
 
 
-def prompt(from_object, to_object, content):
+def execute_prompt(from_object, to_object, content):
     """ Executes the prompt and returns the result. """
 
     # If object implements the prompt method, use it
@@ -171,23 +183,26 @@ def on(object):
 
 def name_to_object(name):
     """ Returns the object with the given name. Returns None if the object is not registered. """
+
+    # TODO: correct support of execution counts
+    name = name.split('[', maxsplit=1)[0]
     return object_names.get(name, None)
 
-def name_to_actor(name):
+def name_to_default_actor(name):
     """ Returns the actor name."""
     object = name_to_object(name)
     if id(object) not in magic_objects:
-        raise ValueError(f"Prompting is not turned on for {object}.")
+        raise ValueError(f"Prompting is not turned on for {name}.")
     
-    actor = magic_objects[id(object)].actor
+    default_actor = magic_objects[id(object)].default_actor
 
-    if id(actor) not in magic_objects:
-        raise ValueError(f"Prompting is not turned on for {actor}.")
+    if id(default_actor) not in magic_objects:
+        raise ValueError(f"Prompting is not turned on for {default_actor}.")
 
-    return magic_objects[id(actor)].name
+    return magic_objects[id(default_actor)].name
 
 
-def turn_on(object, init=None, actor=None, name=None, active=True,
+def turn_on(object, init=None, actor=None, name=None, active=True, default_actor=None,
             steps=None, engine='openai', api_key=None, auto_prompt=False, magic_type=True):
     """
     Activates the connector between the python interpreter and intellegence engine.
@@ -196,6 +211,7 @@ def turn_on(object, init=None, actor=None, name=None, active=True,
     :param actor: Prompting actor that instantiated the object. 
     :param name: The name of the object. If not specified, the name is taken from the object.
     :param active: If True, the object can actively react on the top level thread.
+    :param default_actor: The default actor that is being prompted by undirectional statements. 
     :param steps: Can be None, or a number, including inf. Specifies how many cells the object can create and run at a time.
     :param engine: Can be None, 'openai', 'prompting'
     :param api_key: Can be None, or the API key for the engine.
@@ -238,6 +254,7 @@ def turn_on(object, init=None, actor=None, name=None, active=True,
     I.about = about
     I.init = init
     I.actor = actor
+    I.default_actor = default_actor if default_actor is not None else actor
     I.steps = steps
     I.auto_prompt = auto_prompt
 
@@ -257,6 +274,8 @@ def turn_on(object, init=None, actor=None, name=None, active=True,
         I.engine = EngineEcho(api_key)
     elif engine == 'print':
         I.engine = EnginePrint(api_key)
+    elif engine == 'python':
+        I.engine = EnginePython(api_key)
     else:
         raise ValueError(f"Unknown engine: {engine}")
 
