@@ -4,16 +4,19 @@
 from __future__ import print_function
 
 import IPython.core.magic
+from IPython.core.ultratb import AutoFormattedTB
 from IPython.display import display, Javascript, Markdown, Code
+from IPython.utils.capture import capture_output
 from IPython.core.magic import (Magics, magics_class, line_magic,
                                 cell_magic, line_cell_magic)
-import ast      # AST is also magic, right?
-from parsimonious.nodes import NodeVisitor      # And so are PEGs!
+
+from parsimonious.nodes import NodeVisitor
 from parsimonious.grammar import Grammar
+from parsimonious.exceptions import ParseError
 
 import openai, os, getpass, json
 from notebook.utils import to_api_path
-from .engine import on, turn_on, prompt, name_to_object, name_to_actor
+from .engine import on, turn_on, execute_prompt, name_to_object, name_to_default_actor
 
 @magics_class
 class KeyMagics(Magics):
@@ -22,7 +25,7 @@ class KeyMagics(Magics):
     def __init__(self, shell, **kwargs):
         super().__init__(shell, **kwargs)   
         self.shell = shell
-        shell.events.register('post_run_cell', self.post_run_cell)
+        #shell.events.register('post_run_cell', self.post_run_cell)
 
         self.name = "Arthur"            # Default Arthur-type AGI name
         self.actor = "User"             # Default user name
@@ -48,9 +51,19 @@ class KeyMagics(Magics):
 
     # TODO: change the prompt to use the actor name
 
+    @line_magic
+    def dprompt(self, line):
+        "Identifies and executes the prompt for: From: prompt"
+        from_, content = line.split(maxsplit = 1)      # @object Prompt 
+        to_object = name_to_default_actor(from_)
+        line = f"{from_} {to_object} {content}"
+        return self.prompt(line)
+
+
+
     @line_magic
     def prompt(self, line):
-        "Identifies and executes the prompt for: @object prompt"
+        "Identifies and executes the prompt for: From: @To prompt"
         from_, to, content = line.split(maxsplit = 2)      # @object Prompt 
         #print(from_)
         #print(to)
@@ -61,10 +74,23 @@ class KeyMagics(Magics):
         to_object = name_to_object(to)
         if from_object and to_object:
             # TODO: should we actually be using jupyter_client?
-            code = prompt(from_object, to_object, content)
-            if code:
-                add_code_cell(code, execute = False)
-                self.watched[self.shell.execution_count + 1] = from_object
+            new_prompt = execute_prompt(from_object, to_object, content)
+            if new_prompt:
+                try:
+                    prompt_to_python(new_prompt.split('\n'))
+                except ParseError as e:
+                    code = True
+                    display(Markdown(new_prompt))
+                    error = f"Syntax error in the response above. {e}. Please, fix the response and try again."
+                    # Sanitize the error message, grammar doesn't allow @ and # in the object prompt
+                    error = error.replace('@', '<at>').replace('#', '<hash>')
+                    new_prompt = f"Error: @{to} {error}"
+                    print(f"{new_prompt}")
+
+                add_code_cell(new_prompt, execute = False)
+                #self.watched[self.shell.execution_count + 1] = from_object
+            else:
+                print(f"Empty prompt returned by {to}")
             #print(response, from_, to)
 
             #if request:
@@ -83,32 +109,78 @@ class KeyMagics(Magics):
 
         # return Markdown(input)
 
-    def post_run_cell(self, result):
-        pass
+        return 2345
 
-    def post_run_cell(self, result):
-        if result.execution_count in self.watched:
 
-            self.watched.pop(result.execution_count)
+    @line_magic
+    def my_magic(self, line):
+        display(Markdown(f'***\n```python\ncode\n```\n***'))
+        return "Test"
 
-            formatted_cell = f'In [{result.execution_count}]: {result.info.raw_cell}'
-            if result.result is not None:
-                formatted_cell += f'\nOut [{result.execution_count}]: {result.result}'
-            elif result.error_in_exec is not None:
-                from IPython.core.ultratb import AutoFormattedTB
-                color_tb = AutoFormattedTB(mode = 'Plain', color_scheme = 'NoColor', tb_offset = 1)
-                tb_list = color_tb.structured_traceback(etype = type(result.error_in_exec), value = result.error_in_exec, tb = result.error_in_exec.__traceback__)
-                tb_str = '\n'.join(tb_list)
+    @line_magic
+    def pprompt(self, line):
+        from_, code = line.split(maxsplit = 1)      # @object code
+
+        display(Markdown(f'***\n```python\n{code}\n```\n***'))
+
+        with capture_output() as captured:
+            result = self.shell.run_cell(code) #, store_history=False)
+
+        new_prompt = f"Out[{self.shell.execution_count}]: @{from_} {captured.outputs[0].data['text/plain']}"
+        #if result.result is not None:
+        #    new_prompt += f'{result.result}'
+
+        if result.error_before_exec is not None or result.error_in_exec is not None:
+            exception = result.error_before_exec or result.error_in_exec
+
+            # Create an instance of the InteractiveTB class to format the traceback
+            tb_formatter = AutoFormattedTB(mode='Context', tb_offset=0)
+            traces = type(exception), exception, exception.__traceback__
+            formatted_traceback = tb_formatter.text(*traces)
+            new_prompt += f'\n\nError: @{from_} {formatted_traceback}'
+
+
+        # Add the captured output, stdout, and stderr
+        if captured.stderr:
+            new_prompt += f'\n\nStderr: @{from_} {captured.stderr}'
+            display(Markdown(f'```\n{captured.stderr}\n```'))
+
+        if captured.stdout:
+            new_prompt += f'\n\nStdout: @{from_} {captured.stdout}'
+            display(Markdown(f'```\n{captured.stdout}\n```'))
+
+        if new_prompt:
+            add_code_cell(new_prompt, execute = False)
+
+
+        #for output in captured.outputs:
+        #    display(output)
+
+        #print('pprompt', line, output, stdout, stderr)
+
+
+    # def post_run_cell(self, result):
+    #     if result.execution_count in self.watched:
+
+    #         self.watched.pop(result.execution_count)
+
+    #         formatted_cell = f'In [{result.execution_count}]: {result.info.raw_cell}'
+    #         if result.result is not None:
+    #             formatted_cell += f'\nOut [{result.execution_count}]: {result.result}'
+    #         elif result.error_in_exec is not None:
+    #             color_tb = AutoFormattedTB(mode = 'Plain', color_scheme = 'NoColor', tb_offset = 1)
+    #             tb_list = color_tb.structured_traceback(etype = type(result.error_in_exec), value = result.error_in_exec, tb = result.error_in_exec.__traceback__)
+    #             tb_str = '\n'.join(tb_list)
                 
-                #error_type = type(result.error_in_exec).__name__
-                #formatted_cell += f'\n{error_type}: {result.error_in_exec}'
-                formatted_cell += f'\n{tb_str}'
-            elif result.error_before_exec is not None:
-                error_type = type(result.error_before_exec).__name__
-                formatted_cell += f'\n{error_type}: {result.error_before_exec.msg} in {result.error_before_exec.text}'
+    #             #error_type = type(result.error_in_exec).__name__
+    #             #formatted_cell += f'\n{error_type}: {result.error_in_exec}'
+    #             formatted_cell += f'\n{tb_str}'
+    #         elif result.error_before_exec is not None:
+    #             error_type = type(result.error_before_exec).__name__
+    #             formatted_cell += f'\n{error_type}: {result.error_before_exec.msg} in {result.error_before_exec.text}'
                         
-            code = f"@{self.name}\n" + formatted_cell
-            add_code_cell(code, execute = False)
+    #         code = f"@{self.name}\n" + formatted_cell
+    #         add_code_cell(code, execute = False)
 
 
 
@@ -210,70 +282,116 @@ class KeyMagics(Magics):
 # https://ipython.readthedocs.io/en/stable/config/inputtransforms.html
 
 
-# Grammar for LLM interface, FIXME: content can contain ':' for now
-arthur_grammar = Grammar(
+# Grammar for LLM/User interface, FIXME: content can contain ':' for now
+prompt_grammar = Grammar(
    r"""
-    default_rule = (ws / multi_line_code / inline_code / prompt / hashtag)+
-    
-    multi_line_code = call "```" python? code "```" asterisk?
-    inline_code = call "`" code "`" asterisk?
+    grammar = from prompts
+
+    from = object ":" ws
+    acting_as = object ":" ws
+    acting_stop = newline newline
+
+    prompts = (ws / prompt / default_prompt)+
+    prompt = "@" (code_prompt / object_prompt)
+    default_prompt = content / ""
+    object_prompt = object asterisk? ws content
+
+    code_prompt = (multi_line_code / inline_code) asterisk?
+    multi_line_code = "```" python? code "```" 
+    inline_code = "`" python? code "`"
     python = "python" ws
     code = ~r"([^`]+)"
-    
-    prompt = from? to? help? asterisk* ws content
- 
-    call = "@" 
-    from = object ":"
-    to = ws? call object
-    
-    hashtag = "#" object help? asterisk? previous? 
+
     
     asterisk = "*"
+    newline = ~"\n"
     help = "?"
     previous = "^"
-    object = ~r"[0-9A-z_.]+"
+    execution_number = "[" ~r"[0-9]+" "]"
+
+    object = ~r"[0-9A-z_.]+" execution_number?
     ws = ~r"\s+"i 
 
-    content = ~r"([^#@]+)"
+    content = (acting_as / acting_stop / multi_line_code / inline_code / hashtag / newline / text)+
+    text = ~r"([^#@`\n]+)"
+
+    hashtag = "#" object help? asterisk? previous? 
     """
 )
 
-class ArthurVisitor(NodeVisitor):
+
+
+class PromptVisitor(NodeVisitor):
     def __init__(self):
         self.code_lines = []
+        self.acting_as = None
+        self.from_ = None
                 
     def visit_asterisk(self, node, visited_children):
         pass
 
     def visit_help(self, node, visited_children):
         pass
-    
-    def visit_code(self, node, visited_children):
+
+
+    def visit_default_prompt(self, node, visited_children):
+
+        # This is an alias for the content node
+        content, = node.children
+        from_object = self.from_
+        line = f'%dprompt {from_object} {content.text}'
+        self.code_lines.append(line + '\n')
+
+
+        # Get the list of hashtags. If any, add %hashtag line
+        hashtags = ' '.join([child.text.strip() for child in content.children if child.expr_name == 'hashtag'])
+        if hashtags:
+            line = f'%hashtag {from_object} {hashtags}'
+            self.code_lines.append(line + '\n')
+            
+
+        # TODO: analyze children, there can be code blocks, hashtags, etc.
+        if '`' in content.text:
+            # get snippet around the '`'            
+            left,right = content.text.split('`', maxsplit = 1)
+            if not left.endswith('@'):
+                snippet = left[-5:] + right[:20]
+                snippet = snippet.replace('`', '')
+
+                warning = f"Encountered code block without an execution intent, context: {snippet}."
+                print("Warning: {warning}")
+
+                line = f'%prompt Warning {from_object} {warning}'
+                self.code_lines.append(line + '\n')
+
+            
+
+    def visit_code_prompt(self, node, visited_children):
+        python_code, asterisk = visited_children
+        _,python,code,_ = python_code.children[0]
+        #python,code = python_code[0]
         # ast.parse(node.text.split("\n"))
-        self.code_lines.extend([line + '\n' for line in node.text.split('\n')])
-    
-    def visit_prompt(self, node, visited_children):
-        from_,to,help,asterisk,_,content = visited_children
+        from_object = self.from_
+        self.code_lines.append(f'%pprompt {from_object} {code.text}\n')
+        
+    def visit_from(self, node, visited_children):
+        self.from_= visited_children[0].text
+        self.acting_as = None
 
-        declared = lambda x: x.children
+    def visit_acting_as(self, node, visited_children):
+        self.acting_as = visited_children[0].text
+        #print(f'Acting as {self.acting_as}')
 
-        if not declared(from_) and not declared(to):
-            raise Exception("Prompt must have a From: or @To.")            
+    def visit_acting_stop(self, node, visited_children):
+        self.acting_as = None
+        #print(f'Acting as {self.acting_as}')
 
-        # Initialize from_object and to_object
-        if declared(from_):
-            from_object = from_.children[0].text[:-1]
 
-        if declared(to):
-            _,call,object = to.children[0]
-            to_object = object.text
-
-        # Fill in the defaults, if not available
-        if not declared(from_):
-            from_object = name_to_actor(to_object)
-
-        if not declared(to):
-            to_object = name_to_actor(from_object)
+    def visit_object_prompt(self, node, visited_children):
+        to,asterisk,_,content = visited_children
+        #content_text, hashtag = content
+        from_object = self.from_
+        to_object = to.text
 
         line = f'%prompt {from_object} {to_object} {content.text}'
         self.code_lines.append(line + '\n')
@@ -286,26 +404,31 @@ class ArthurVisitor(NodeVisitor):
         return node # visited_children or node
 
 
-def arthur_to_python(lines):
+def prompt_to_python(lines):
     """
         This transforms lines from @```python.code()``` to python.code()
         and from @object Prompt to %prompt object.__prompt__("Prompt").
         This also processes #hastag tags, replacing it with %memory
+    """    
+    tree = prompt_grammar.parse('\n'.join(lines))
+    visitor = PromptVisitor()
+    visitor.visit(tree)
+    return visitor.code_lines
+
+
+def preprocess_cell(lines):
+    """
+        This is the main entry point for the input transformation.
     """
 
-    # check if lines is valid Arthur prompt
+    # check if lines is a valid prompt, if not assume this is a python code
     try:
-        tree = arthur_grammar.parse('\n'.join(lines))
-        visitor = ArthurVisitor()
-        visitor.visit(tree)
-        return visitor.code_lines
+        return prompt_to_python(lines)
     except:
         return lines
 
 
 
-
-
 def add_response_cell(markdown):
     "Adds a new markdown cell below the current cell"
 
@@ -450,5 +573,5 @@ def load_ipython_extension(ipython):
     ipython.register_magics(magics)
 
     # Add as first element of the list of input transformers
-    ipython.input_transformers_cleanup.insert(0, arthur_to_python)
+    ipython.input_transformers_cleanup.insert(0, preprocess_cell)
 
